import ipaddr
from netaddr import IPNetwork
from ipaddress import IPv6Network, IPv6Address
from random import getrandbits
import sys


def get_macs(mac, count, offset=1):
    """
    Take mac as start mac returns the count of macs in a list
    """
    mac_list = list()
    for i in range(count):
        mac_address = "{:012X}".format(int(mac, 16) + offset * i)
        mac_address = ":".join(
            format(s, "02x") for s in bytearray.fromhex(mac_address)
        )
        mac_list.append(mac_address)
    return mac_list


def get_ip_addresses(ip, count):
    """
    Take ip as start ip returns the count of ips in a list
    """
    ip_list = list()
    for i in range(count):
        ipv4address = ipaddr.IPv4Address(ip)
        ipv4address = ipv4address + i
        value = ipv4address._string_from_ip_int(ipv4address._ip)
        ip_list.append(value)
    return ip_list


def get_ipv6_addrs(ip, count):
    """
    Get N IPv6 addresses in a subnet.
    Args:
        subnet (str): IPv6 subnet, e.g., '2001::1/64'
        number_of_ip (int): Number of IP addresses to get
    Return:
        Return n IPv6 addresses in this subnet in a list.
    """
    subnet = str(IPNetwork(ip).network) + "/" + str(ip.split("/")[1])
    if sys.version_info[0] == 2:
        subnet = unicode(subnet, "utf-8")
    ipv6_list = []
    for i in range(count):
        network = IPv6Network(subnet)
        address = IPv6Address(
            network.network_address
            + getrandbits(network.max_prefixlen - network.prefixlen)
        )
        ipv6_list.append(str(address))

    return ipv6_list


def stats_ok(api, packets, utils):
    """
    Returns true if stats are as expected, false otherwise.
    """
    _, flow_stats = utils.get_all_stats(api)

    flow_rx = sum([f.frames_rx for f in flow_stats])
    return flow_rx == packets


def test_compact(api, utils):
    """
    1. Configure 1000 similar devices contain Ether>IPv4>BGPv4
       route in two ports
    2. Create three traffic each traffic with diff set of endpoints
    3. Concrete class should compact the devices to IxNet DeviceGroup.
    4. Validate all through RestPy
    """
    SIZE = 1024
    PACKETS = 1000

    config_values = dict()
    num_of_devices = 1000
    num_of_routes = 1000
    config_values["d1_name"] = "Tx Device 1"
    config_values["d2_name"] = "Rx Device 1"
    config_values["d3_name"] = "Rx Device 3"
    config_values["Multiplier"] = num_of_devices
    rx_device_with_rr = 3

    config = api.config()
    api.enable_scaling(True)

    tx_port, rx_port = config.ports.port(
        name="Tx Port", location=utils.settings.ports[0]
    ).port(name="Rx Port", location=utils.settings.ports[1])

    l1 = config.layer1.layer1()[0]
    l1.name = "l1"
    l1.port_names = [rx_port.name, tx_port.name]
    l1.media = utils.settings.media
    l1.speed = utils.settings.speed

    macs = get_macs("000000000011", 2 * num_of_devices)

    config_values["tx_macs"], config_values["rx_macs"] = macs[::2], macs[1::2]
    config_values["vlan_ids"] = [str(i) for i in range(1, num_of_devices + 1)]

    ip_adds = get_ip_addresses("10.10.2.1", 2 * num_of_devices)
    ipv6_adds = get_ipv6_addrs("2001::1/64", 2 * num_of_devices)

    config_values["tx_adds"], config_values["rx_adds"] = (
        ip_adds[::2],
        ip_adds[1::2],
    )

    config_values["tx_ipv6_adds"], config_values["rx_ipv6_adds"] = (
        ipv6_adds[::2],
        ipv6_adds[1::2],
    )
    config_values["tx_rr_add1"] = get_ip_addresses("200.1.0.0", num_of_routes)
    config_values["tx_rr_add2"] = get_ip_addresses("201.1.0.0", num_of_routes)
    next_hop_addr = get_ip_addresses("4.4.4.1", num_of_routes)

    config_values["rx_rr_add1"] = "210.1.0.0"

    for i in range(1, num_of_devices + 1):
        tx_device = config.devices.add()
        tx_device.name = "Tx Device {0}".format(i)
        tx_eth = tx_device.ethernets.add()
        tx_eth.connection.port_name = tx_port.name
        tx_eth.name = "Tx eth {0}".format(i)
        tx_eth.mac = config_values["tx_macs"][i - 1]
        tx_vlan = tx_eth.vlans.vlan()[-1]
        tx_vlan.name = "Tx vlan {0}".format(i)
        tx_vlan.id = int(config_values["vlan_ids"][i - 1])
        tx_ip = tx_eth.ipv4_addresses.add()
        tx_ip.name = "Tx IP {0}".format(i)
        tx_ip.address = config_values["tx_adds"][i - 1]
        tx_ip.gateway = config_values["rx_adds"][i - 1]
        tx_ip.prefix = 24

        tx_ipv6 = tx_eth.ipv6_addresses.add()
        tx_ipv6.name = "Tx IP v6{0}".format(i)
        tx_ipv6.address = config_values["tx_ipv6_adds"][i - 1]
        tx_ipv6.gateway = config_values["rx_ipv6_adds"][i - 1]
        tx_ipv6.prefix = 64

        tx_bgp = tx_device.bgp
        tx_bgp.router_id = config_values["tx_adds"][i - 1]
        tx_bgp_int = tx_bgp.ipv4_interfaces.add()
        tx_bgp_int.ipv4_name = tx_ip.name
        tx_peer = tx_bgp_int.peers.add()
        tx_peer.name = "BGP Peer {0}".format(i)
        tx_peer.as_type = "ibgp"
        tx_peer.peer_address = config_values["rx_adds"][i - 1]
        tx_peer.as_number = 65200

        tx_rr = tx_peer.v4_routes.add(name="Tx RR {0}".format(i))
        tx_rr.addresses.add(
            count=20, address=config_values["tx_rr_add1"][i - 1], prefix=32
        )
        tx_rr.addresses.add(
            count=10, address=config_values["tx_rr_add2"][i - 1], prefix=24
        )
        tx_rr.next_hop_ipv4_address = next_hop_addr[i - 1]

    for i in range(1, num_of_devices + 1):
        rx_device = config.devices.add()
        rx_device.name = "Rx Device {0}".format(i)
        rx_eth = rx_device.ethernets.add()
        rx_eth.connection.port_name = rx_port.name
        rx_eth.name = "Rx eth {0}".format(i)
        rx_eth.mac = config_values["rx_macs"][i - 1]
        rx_vlan = rx_eth.vlans.vlan()[-1]
        rx_vlan.name = "Rx vlan {0}".format(i)
        rx_vlan.id = int(config_values["vlan_ids"][i - 1])
        rx_ip = rx_eth.ipv4_addresses.add()
        rx_ip.name = "Rx IP {0}".format(i)
        rx_ip.address = config_values["rx_adds"][i - 1]
        rx_ip.gateway = config_values["tx_adds"][i - 1]
        rx_ip.prefix = 24

        rx_ipv6 = rx_eth.ipv6_addresses.add()
        rx_ipv6.name = "Rx IP v6{0}".format(i)
        rx_ipv6.address = config_values["rx_ipv6_adds"][i - 1]
        rx_ipv6.gateway = config_values["tx_ipv6_adds"][i - 1]
        rx_ipv6.prefix = 64

        rx_bgp = rx_device.bgp
        rx_bgp.router_id = config_values["rx_adds"][i - 1]
        rx_bgp_int = rx_bgp.ipv4_interfaces.add()
        rx_bgp_int.ipv4_name = rx_ip.name
        rx_peer = rx_bgp_int.peers.add()
        rx_peer.name = "Rx Bgp {0}".format(i)
        rx_peer.as_type = "ibgp"
        rx_peer.peer_address = config_values["tx_adds"][i - 1]
        rx_peer.as_number = 65200


        if i == rx_device_with_rr:
            rx_rr = rx_peer.v4_routes.add(name="Rx RR {0}".format(i))
            rx_rr.addresses.add(
                count=1000,
                address=config_values["rx_rr_add1"],
                prefix=32,
            )

    config_values["f1_name"] = "f1"
    config_values["f2_name"] = "f2"
    config_values["f3_name"] = "f3"
    flow1, flow2, flow3 = (
        config.flows.flow(name=config_values["f1_name"])
        .flow(name=config_values["f2_name"])
        .flow(name=config_values["f3_name"])
    )

    # Route Range as endpoints
    flow1.tx_rx.device.tx_names = ["Tx RR 2", "Tx RR 4"]
    flow1.tx_rx.device.rx_names = ["Rx RR 3"]
    _, _, tcp = flow1.packet.ethernet().ipv4().tcp()
    tcp.src_port.value = 555
    tcp.dst_port.value = 666

    flow1.size.fixed = SIZE
    flow1.duration.fixed_packets.packets = PACKETS

    flow1.metrics.enable = True
    flow1.metrics.loss = True

    # Ethernet as endpoints
    flow2.tx_rx.device.tx_names = ["Tx eth 2"]
    flow2.tx_rx.device.rx_names = ["Rx eth 2"]

    flow2.size.fixed = SIZE
    flow2.duration.fixed_packets.packets = PACKETS

    flow2.metrics.enable = True
    flow2.metrics.loss = True

    # All Ipv4 as endpoints
    flow3.tx_rx.device.tx_names = [
        "Tx IP {0}".format(i) for i in range(1, num_of_devices + 1)
    ]
    flow3.tx_rx.device.rx_names = [
        "Rx IP {0}".format(i) for i in range(1, num_of_devices + 1)
    ]

    flow3.size.fixed = SIZE
    flow3.duration.fixed_packets.packets = PACKETS

    flow3.metrics.enable = True
    flow3.metrics.loss = True

    api.set_config(config)

    validate_compact_config(api, config_values, rx_device_with_rr)
    print("Starting all protocols ...")
    # ps = api.protocol_state()
    # ps.state = ps.START
    # api.set_protocol_state(ps)

    cs = api.control_state()
    cs.protocol.all.state = cs.protocol.all.START
    api.set_control_state(cs)

    print("Starting transmit on all flows ...")

    cs = api.control_state()
    cs.traffic.flow_transmit.state = cs.traffic.flow_transmit.START
    api.set_control_state(cs)

    # utils.start_traffic(api, config, start_capture=False)
    utils.wait_for(
        lambda: stats_ok(api, PACKETS * 3, utils), "stats to be as expected"
    )

    cs = api.control_state()
    cs.protocol.route.state = cs.protocol.route.WITHDRAW
    cs.protocol.route.names = ["Tx RR 4", "Rx RR 3"]
    api.set_control_state(cs)


    validate_route_withdraw(api, config_values)


def compare(actual, expected):
    return all([a == b for a, b in zip(actual, expected)])


def validate_compact_config(api, config_values, rx_device_with_rr):
    """
    Validate attributes using RestPy
    """
    ixnetwork = api._ixnetwork

    d1 = ixnetwork.Topology.find().DeviceGroup.find(
        Name=config_values["d1_name"]
    )
    d2 = ixnetwork.Topology.find().DeviceGroup.find(
        Name=config_values["d2_name"]
    )
    d3 = ixnetwork.Topology.find().DeviceGroup.find(
        Name=config_values["d3_name"]
    )

    assert d1.Multiplier == config_values["Multiplier"]
    # assert Macs
    d3_mac = config_values["rx_macs"].pop(rx_device_with_rr - 1)
    assert compare(d1.Ethernet.find().Mac.Values, config_values["tx_macs"])
    assert compare(d2.Ethernet.find().Mac.Values, config_values["rx_macs"])
    assert d3.Ethernet.find().Mac.Values[0] == d3_mac

    assert compare(
        d1.Ethernet.find().Vlan.find().VlanId.Values, config_values["vlan_ids"]
    )

    # Assert values for d1
    assert compare(
        d1.Ethernet.find().Ipv4.find().Address.Values, config_values["tx_adds"]
    )
    assert compare(
        d1.Ethernet.find().Ipv4.find().GatewayIp.Values,
        config_values["rx_adds"],
    )
    assert compare(
        d1.Ethernet.find().Ipv4.find().BgpIpv4Peer.find().DutIp.Values,
        config_values["rx_adds"],
    )
    assert compare(
        d1.Ethernet.find().Ipv6.find().Address.Values,
        config_values["tx_ipv6_adds"],
    )
    assert compare(
        d1.Ethernet.find().Ipv6.find().GatewayIp.Values,
        config_values["rx_ipv6_adds"],
    )

    # Assert values for d2
    d3_ip = config_values["rx_adds"].pop(rx_device_with_rr - 1)
    d3_gateway = config_values["tx_adds"].pop(rx_device_with_rr - 1)
    d3_ipv6 = config_values["rx_ipv6_adds"].pop(rx_device_with_rr - 1)
    d3_ipv6_gateway = config_values["tx_ipv6_adds"].pop(rx_device_with_rr - 1)
    assert compare(
        d2.Ethernet.find().Ipv4.find().Address.Values, config_values["rx_adds"]
    )
    assert compare(
        d2.Ethernet.find().Ipv4.find().GatewayIp.Values,
        config_values["tx_adds"],
    )
    assert compare(
        d2.Ethernet.find().Ipv4.find().BgpIpv4Peer.find().DutIp.Values,
        config_values["tx_adds"],
    )
    assert compare(
        d2.Ethernet.find().Ipv6.find().Address.Values,
        config_values["rx_ipv6_adds"],
    )
    assert compare(
        d2.Ethernet.find().Ipv6.find().GatewayIp.Values,
        config_values["tx_ipv6_adds"],
    )

    # Assert values for d3
    assert d3.Ethernet.find().Ipv4.find().Address.Values[0] == d3_ip
    assert d3.Ethernet.find().Ipv4.find().GatewayIp.Values[0] == d3_gateway
    assert d3.Ethernet.find().Ipv6.find().Address.Values[0] == d3_ipv6
    assert (
        d3.Ethernet.find().Ipv6.find().GatewayIp.Values[0] == d3_ipv6_gateway
    )
    assert (
        d3.Ethernet.find().Ipv4.find().BgpIpv4Peer.find().DutIp.Values[0]
        == d3_gateway
    )

    # Assert Network Group
    assert compare(
        d1.NetworkGroup.find()
        .Ipv4PrefixPools.find()
        .NetworkAddress.Values[::2],
        config_values["tx_rr_add1"],
    )
    assert compare(
        d1.NetworkGroup.find()
        .Ipv4PrefixPools.find()
        .NetworkAddress.Values[1::2],
        config_values["tx_rr_add2"],
    )
    assert (
        d3.NetworkGroup.find().Ipv4PrefixPools.find().NetworkAddress.Values[0]
        == config_values["rx_rr_add1"]
    )

    # Assert traffic endpoint for f1
    d1_ixn_pool = d1.NetworkGroup.find().Ipv4PrefixPools.find()
    f1_ixn_traffic = ixnetwork.Traffic.TrafficItem.find(
        Name=config_values["f1_name"]
    )
    assert len(f1_ixn_traffic) == 1
    f1_ixn_endpoint = f1_ixn_traffic.EndpointSet.find()
    assert f1_ixn_endpoint.Sources == []
    f1_cfg_endpoint = [
        {"arg1": d1_ixn_pool.href, "arg2": 1, "arg3": 1, "arg4": 3, "arg5": 2},
        {"arg1": d1_ixn_pool.href, "arg2": 1, "arg3": 1, "arg4": 7, "arg5": 2},
    ]
    assert f1_ixn_endpoint.ScalableSources == f1_cfg_endpoint

    # Assert traffic endpoint for f2
    f2_ixn_traffic = ixnetwork.Traffic.TrafficItem.find(
        Name=config_values["f2_name"]
    )
    f2_ixn_endpoint = f2_ixn_traffic.EndpointSet.find()
    f2_cfg_src_endpoint = [
        {
            "arg1": d1.href + "/ethernet/1",
            "arg2": 1,
            "arg3": 1,
            "arg4": 2,
            "arg5": 1,
        }
    ]

    f2_cfg_dst_endpoint = [
        {
            "arg1": d2.href + "/ethernet/1",
            "arg2": 1,
            "arg3": 1,
            "arg4": 2,
            "arg5": 1,
        }
    ]
    assert f2_ixn_endpoint.ScalableSources == f2_cfg_src_endpoint
    assert f2_ixn_endpoint.ScalableDestinations == f2_cfg_dst_endpoint

    # Assert traffic endpoint for f3
    f3_ixn_traffic = ixnetwork.Traffic.TrafficItem.find(
        Name=config_values["f3_name"]
    )
    f3_ixn_endpoint = f3_ixn_traffic.EndpointSet.find()
    assert f3_ixn_endpoint.Sources[0] == d1.href + "/ethernet/1/ipv4/1"


def validate_route_withdraw(api, config_values):
    ixnetwork = api._ixnetwork
    d1 = ixnetwork.Topology.find().DeviceGroup.find(
        Name=config_values["d1_name"]
    )
    d3 = ixnetwork.Topology.find().DeviceGroup.find(
        Name=config_values["d3_name"]
    )
    d1_ixn_pool = d1.NetworkGroup.find().Ipv4PrefixPools.find()
    d3_ixn_pool = d3.NetworkGroup.find().Ipv4PrefixPools.find()
    assert (d1_ixn_pool.BgpIPRouteProperty.find().Active.Values)[6] == "false"
    assert (d1_ixn_pool.BgpIPRouteProperty.find().Active.Values)[7] == "false"
    assert (d3_ixn_pool.BgpIPRouteProperty.find().Active.Values)[0] == "false"


import pytest

if __name__ == "__main__":
    pytest.main(["-s", __file__])



from snappi_ixnetwork.device.base import Base
from snappi_ixnetwork.logger import get_ixnet_logger


class Mka(Base):
    _BASIC = {
        "key_derivation_function": {
            "ixn_attr": "keyDerivationFunction",
            "enum_map": {
                "aes_cmac_128": "aescmac128",
                "aes_cmac_256": "aescmac256",
            },
        },
        "macsec_capability": {
            "ixn_attr": "macsecCapability",
            "enum_map": {
                "macsec_not_implemented": "macsecnotimplemented",
                "macsec_integrity_without_confidentiality": "macsecintegritywithoutconfidentiality",
                "macsec_integrity_with_no_confidentiality_offset": "acsecintegritywithnoconfidentialityoffset",
                "macsec_integrity_with_confidentiality_offset": "macsecintegritywithconfidentialityoffset",
            },
        },
        "actor_priority": "keyServerPriority",
        "macsec_desired": "macsecDesired",
        "mka_version": "mkaVersion",
        "mka_hello_time": "mkaHelloTime",
        "send_icv_indicatior_in_mkpdu": "sendICVIndicator",
        "delay_protect": "delayProtect",
    }

    _KEY_SERVER = {
        "confidentialty_offset": {
            "ixn_attr": "confidentialityOffset",
            "enum_map": {
                "no_confidentiality": "noconfidentiality",
                "no_confidentiality_offset": "noconfidentialityoffset",
                "confidentiality_offset_30_octets": "confidentialityoffset30octets",
                "confidentiality_offset_50_octets": "confidentialityoffset50octets",
            },
        },
        "cipher_suite": {
            "ixn_attr": "cipherSuite",
            "enum_map": {
                "gcm_aes_128": "aes128",
                "gcm_aes_256": "aes256",
                "gcm_aes_xpn_128": "aesxpn128",
                "gcm_aes_xpn_256": "aesxpn256",
            },
        },
        "starting_key_number": "startingKeyNumber",
        "starting_distributed_an": "startingDistributedAN",
        "rekey_threshold_pn": "rekeyThresholdPN",
        "rekey_threshold_xpn": "rekeyThresholdXPN",
    }

    _TXSC = {
        "system_id": "systemId",
        "port_id": "portId",
        "starting_message_number": "startingMessageNumber",
    }

    def __init__(self, ngpf):
        super(Mka, self).__init__()
        self._ngpf = ngpf
        self.logger = get_ixnet_logger(__name__)

    def config(self, device):
        self.logger.debug("Configuring MKA protocol")
        macsec = device.get("macsec")
        if macsec is None:
            return
        self._config_ethernet_interfaces(device)

    def _is_valid(self, ethernet_name, kay):
        is_valid = True
        if kay is None:
            is_valid = False
        else:
            self.logger.debug(
                "Validating KaY of ethernert interface %s" % (ethernet_name)
            )
            # Validate basic properties
            basic = kay.get("basic")
            key_src = basic.key_source
            if key_src.choice == "psk":
                psks = key_src.psks
                if len(psks) == 0:
                    self._ngpf.api.add_error(
                        "No PSK added".format(name=ethernet_name)
                    )
                    is_valid = False
            else:
                self._ngpf.api.add_error(
                    "Key source other than PSK set".format(name=ethernet_name)
                )
                is_valid = False
            # Validate TxSC properties
            tx = kay.get("tx")
            txscs = tx.get("secure_channels")
            if len(txscs) == 0:
                self._ngpf.api.add_error(
                    "No TxSC added".format(name=ethernet_name)
                )
                is_valid = False
            elif len(txscs) > 1:
                self._ngpf.api.add_error(
                    "More than one TxSC added".format(name=ethernet_name)
                )
                is_valid = False
        if is_valid == True:
            self.logger.debug("MKA validation success")
        else:
            self.logger.debug("MKA validation failure")
        return is_valid

    def _config_ethernet_interfaces(self, device):
        self.logger.debug("Configuring MKA interfaces")
        macsec = device.get("macsec")
        ethernet_interfaces = macsec.get("ethernet_interfaces")
        if ethernet_interfaces is None:
            return
        for ethernet_interface in ethernet_interfaces:
            ethernet_name = ethernet_interface.get("eth_name")
            secy = ethernet_interface.get("secure_entity")
            if secy is None:
                continue
            key_generation_protocol = secy.get("key_generation_protocol")
            if key_generation_protocol is None:
                continue
            protocol = key_generation_protocol.choice
            if protocol != "mka":
                continue
            kay = key_generation_protocol.get("mka")
            if kay is None:
                continue
            if not self._is_valid(ethernet_name, kay):
                continue
            self._ngpf.working_dg = self._ngpf.api.ixn_objects.get_working_dg(
                ethernet_name
            )
            self._config_kay(ethernet_interface, kay)

    def _config_kay(self, ethernet_interface, kay):
        self.logger.debug("Configuring KaY")
        ethernet_name = ethernet_interface.get("eth_name")
        ixn_ethernet = self._ngpf.api.ixn_objects.get_object(ethernet_name)
        ixn_mka = self.create_node_elemet(ixn_ethernet, "mka", kay.get("name"))
        self._ngpf.set_device_info(kay, ixn_mka)
        self._config_basic(kay, ixn_mka)
        self._config_txsc(kay, ixn_mka)
        self._config_keyserver(kay, ixn_mka)

    def _config_basic(self, kay, ixn_mka):
        self.logger.debug("Configuring basic properties")
        basic = kay.get("basic")
        self.configure_multivalues(basic, ixn_mka, Mka._BASIC)
        ixn_mka["mkaLifeTime"] = basic.mka_life_time
        ixn_mka["keyType"] = basic.key_source.choice
        self._config_key_source(basic, ixn_mka)
        self._config_rekey_mode(basic, ixn_mka)
        self._config_supported_cipher_suites(basic, ixn_mka)
        self._config_test_start_time(basic, ixn_mka)

    def _is_test_start_time_user_defined(self, macsec):
        self.logger.debug("Checking if MKA is configured")
        ethernet_interfaces = macsec.get("ethernet_interfaces")
        for ethernet_interface in ethernet_interfaces:
            secy = ethernet_interface.get("secure_entity")
            key_generation_protocol = secy.get("key_generation_protocol")
            protocol = key_generation_protocol.choice
            if protocol == "mka":
                kay = key_generation_protocol.get("mka")
                if kay:
                    basic = kay.get("basic")
                    psk_chain_start_time = basic.psk_chain_start_time
                    if (
                        psk_chain_start_time.choice == "utc"
                        and psk_chain_start_time.utc.day is not None
                    ):
                        return True
        return False

    def _clear_overlays_in_globals(self, macsec):
        if self._is_test_start_time_user_defined(macsec):
            ixn_mka_globals_port_settings = (
                self._ngpf._ixnetwork.Globals.Topology.find().Mka.find()
            )
            ixn_mka_globals_port_settings.TestStartTime.ClearOverlays()

    def _config_test_start_time(self, basic, ixn_mka):
        self.logger.debug("Configuring test start time")
        psk_chain_start_time = basic.psk_chain_start_time
        if (
            psk_chain_start_time.choice == "utc"
            and psk_chain_start_time.utc.day is not None
        ):
            utc_day = str(psk_chain_start_time.utc.day)
            utc_month = str(psk_chain_start_time.utc.month)
            utc_year = str(psk_chain_start_time.utc.year)
            utc_hour = str(psk_chain_start_time.utc.hour)
            utc_minute = str(psk_chain_start_time.utc.minute)
            utc_second = str(psk_chain_start_time.utc.second)

            utc_time = (
                utc_day
                + "-"
                + utc_month
                + "-"
                + utc_year
                + " "
                + utc_hour
                + ":"
                + utc_minute
                + ":"
                + utc_second
            )

            ixn_topology = self._ngpf.api._ixnetwork.Globals.Topology.refresh()
            ixn_mka_global_port_settings = ixn_topology.find().Mka.find()
            ixn_mka_global_test_start_time = (
                ixn_mka_global_port_settings.TestStartTime
            )
            ixn_mka_global_test_start_time.ValueList(
                [utc_time] * ixn_mka_global_test_start_time.Count
            )
            self.logger.debug(
                "MKA global per port test start time set to %s"
                % ixn_mka_global_test_start_time
            )

    def _config_supported_cipher_suites(self, basic, ixn_mka):
        self.logger.debug(
            "Configuring basic properties: supported cipher suites"
        )
        supported_cipher_suites = "selectciphers"
        if basic.supported_cipher_suites.gcm_aes_128:
            supported_cipher_suites += " gcm_aes_128"
        if basic.supported_cipher_suites.gcm_aes_256:
            supported_cipher_suites += " gcm_aes_256"
        if basic.supported_cipher_suites.gcm_aes_xpn_128:
            supported_cipher_suites += " gcm_aes_xpn_128"
        if basic.supported_cipher_suites.gcm_aes_xpn_256:
            supported_cipher_suites += " gcm_aes_xpn_256"
        ixn_mka["supportedCipherSuites"] = supported_cipher_suites

    def _config_key_source(self, basic, ixn_mka):
        self.logger.debug("Configuring key source settings")
        key_src = basic.key_source
        if key_src.choice == "psk":
            psks = key_src.psks
            ixn_mka["cakCount"] = len(psks)
            cak_names = []
            cak_values = []
            cak_start_offset_times = []
            cak_lifetime_validities = []
            cak_durations = []
            overlapping_keys = False
            prev_end_offset_time_mm = 0
            for psk in psks:
                cak_names.append(psk.cak_name)
                cak_values.append(psk.cak_value)

                start_offset_time_hh = psk.start_offset_time.hh
                start_offset_time_mm = psk.start_offset_time.mm
                start_offset_time = (
                    str(start_offset_time_hh) + ":" + str(start_offset_time_mm)
                )
                cak_start_offset_times.append(start_offset_time)
                start_offset_time_mm = (
                    start_offset_time_hh * 60 + start_offset_time_mm
                )

                end_offset_time_hh = psk.end_offset_time.hh
                end_offset_time_mm = psk.end_offset_time.mm
                end_offset_time_mm = (
                    end_offset_time_hh * 60 + end_offset_time_mm
                )

                if not overlapping_keys:
                    if start_offset_time_mm < prev_end_offset_time_mm:
                        overlapping_keys = True
                    else:
                        prev_end_offset_time_mm = end_offset_time_mm
                duration_mm = 0
                if end_offset_time_mm == 0:
                    cak_lifetime_validities.append(True)
                else:
                    cak_lifetime_validities.append(False)
                    if end_offset_time_mm > start_offset_time_mm:
                        duration_mm = end_offset_time_mm - start_offset_time_mm
                duration_hh = int(duration_mm / 60)
                duration_mm = duration_mm % 60
                cak_duration = str(duration_hh) + ":" + str(duration_mm)
                cak_durations.append(cak_duration)

            ixn_psk = self.create_node_elemet(ixn_mka, "cakCache", name=None)
            key_derivation_function = basic.key_derivation_function
            ixn_psk["cakName"] = self.multivalue(cak_names)
            if key_derivation_function == "aes_cmac_128":
                ixn_psk["cakValue128"] = self.multivalue(cak_values)
            elif key_derivation_function == "aes_cmac_256":
                ixn_psk["cakValue256"] = self.multivalue(cak_values)
            ixn_psk["keyStartTime"] = self.multivalue(cak_start_offset_times)
            ixn_psk["lifetimeValidity"] = self.multivalue(
                cak_lifetime_validities
            )
            ixn_psk["overlappingKeys"] = overlapping_keys
            ixn_psk["keyDuration"] = self.multivalue(cak_durations)

    def _config_rekey_mode(self, basic, ixn_mka):
        self.logger.debug("Configuring rekey settings")
        rekey_mode = basic.rekey_mode
        if rekey_mode.choice == "dont_rekey":
            ixn_mka["rekeyMode"] = "timerBased"
            ixn_mka["rekeyBehaviour"] = "dontRekey"
        elif rekey_mode.choice == "timer_based":
            timer_based = rekey_mode.timer_based
            ixn_mka["rekeyMode"] = "timerBased"
            if timer_based.choice == "fixed_count":
                ixn_mka["rekeyBehaviour"] = "rekeyFixedCount"
                ixn_mka["periodicRekeyAttempts"] = timer_based.fixed_count
                ixn_mka["periodicRekeyInterval"] = timer_based.interval
            elif timer_based.choice == "continuous":
                ixn_mka["rekeyBehaviour"] = "rekeyContinuous"
                ixn_mka["periodicRekeyInterval"] = timer_based.interval
        elif rekey_mode.choice == "pn_based":
            ixn_mka["rekeyMode"] = "pNBased"
            ixn_mka["rekeyBehaviour"] = "rekeyContinuous"

    def _config_txsc(self, kay, ixn_mka):
        self.logger.debug("Configuring TxSC")
        tx = kay.get("tx")
        txscs = tx.get("secure_channels")
        txsc = txscs[0]
        ixn_txsc = self.create_node_elemet(
            ixn_mka, "txChannels", txsc.get("name")
        )
        self.logger.debug("ixn_txsc %s" % ixn_txsc)
        self.configure_multivalues(txsc, ixn_txsc, Mka._TXSC)

    def _config_keyserver(self, kay, ixn_mka):
        self.logger.debug("Configuring key server")
        key_server = kay.get("key_server")
        if key_server is None:
            return
        self.configure_multivalues(key_server, ixn_mka, Mka._KEY_SERVER)



import json
import time
import io
from snappi_ixnetwork.timer import Timer
from snappi_ixnetwork.logger import get_ixnet_logger


class CommonField:
    def __init__(self, field, offset):
        self.field = field
        self.offset = offset


class Capture(object):
    """Transforms OpenAPI objects into IxNetwork objects

    Args
    ----
    - ixnetworkapi (Api): instance of the Api class

    Transformations
    ---------------
    - /components/schemas/Capture to /vport/capture

    Process
    -------
    - Configure capture according to Filter

    Notes
    -----

    """

    def __init__(self, ixnetworkapi):
        self._api = ixnetworkapi
        self._capture_request = None
        self.logger = get_ixnet_logger(__name__)

    def _import(self, imports):
        if len(imports) > 0:
            errata = self._resource_manager.ImportConfig(
                json.dumps(imports), False
            )
            for item in errata:
                self._api.warning(item)
            return len(errata) == 0
        return True

    def config(self):
        """Overwrite any capture settings"""
        self._resource_manager = self._api._ixnetwork.ResourceManager
        imports = []
        vports = self._api.select_vports()
        for vport in vports.values():
            if (
                vport["capture"]["hardwareEnabled"] is True
                or vport["capture"]["softwareEnabled"] is True
            ):
                capture = {
                    "xpath": vport["capture"]["xpath"],
                    "captureMode": "captureTriggerMode",
                    "hardwareEnabled": False,
                    "softwareEnabled": False,
                }
                imports.append(capture)
        for capture_item in self._api.snappi_config.captures:
            if capture_item.format == "pcap":
                self._api.warning(
                    "pcap format is not supported for IxNetwork, setting capture format to pcapng"
                )
                capture_item.format = "pcapng"
            if capture_item.port_names is None:
                continue
            for port_name in capture_item.port_names:
                capture_mode = "captureTriggerMode"
                if capture_item.get("overwrite"):
                    capture_mode = "captureContinuousMode"
                    reset = {
                        "xpath": vports[port_name]["xpath"]
                        + "/capture/trigger"
                    }
                    self._import(reset)
                capture = {
                    "xpath": vports[port_name]["xpath"] + "/capture",
                    "captureMode": capture_mode,
                    "hardwareEnabled": True,
                    "softwareEnabled": True,
                }
                pallette = {"xpath": capture["xpath"] + "/filterPallette"}
                filter = {"xpath": capture["xpath"] + "/filter"}
                trigger = {"xpath": capture["xpath"] + "/trigger"}
                if len(capture_item.filters) > 0:
                    filter["captureFilterEnable"] = True
                    trigger["captureTriggerEnable"] = True
                    filter["captureFilterFrameSizeEnable"] = False
                    trigger["captureTriggerFrameSizeEnable"] = False
                    filter["captureFilterFrameSizeTo"] = 9198
                    trigger["captureTriggerFrameSizeTo"] = 9198
                    filter["captureFilterFrameSizeFrom"] = 46
                    trigger["captureTriggerFrameSizeFrom"] = 46
                    generic_size = 0
                    expression_string = ""
                    for cap_filter in capture_item.filters:
                        if cap_filter.parent.choice == "ethernet":
                            expression_string = self._config_ethernet_pallette(
                                cap_filter,
                                pallette,
                                trigger,
                                filter,
                                generic_size,
                                expression_string,
                            )
                            generic_size += 14
                        elif cap_filter.parent.choice == "custom":
                            generic_size, expression_string = (
                                self._config_custom_pallete(
                                    cap_filter,
                                    pallette,
                                    trigger,
                                    filter,
                                    generic_size,
                                    expression_string,
                                )
                            )
                        elif cap_filter.parent.choice == "vlan":
                            expression_string = self._config_vlan_pallette(
                                cap_filter,
                                pallette,
                                trigger,
                                filter,
                                generic_size,
                                expression_string,
                            )
                            generic_size += 4
                        elif cap_filter.parent.choice == "ipv4":
                            expression_string = self._config_ipv4_pallete(
                                cap_filter,
                                pallette,
                                trigger,
                                filter,
                                generic_size,
                                expression_string,
                            )
                            generic_size += 20
                        elif cap_filter.parent.choice == "ipv6":
                            expression_string = self._config_ipv6_pallete(
                                cap_filter,
                                pallette,
                                trigger,
                                filter,
                                generic_size,
                                expression_string,
                            )
                            generic_size += 40
                        else:
                            expression_string = self._config_missing_pallete(
                                cap_filter,
                                pallette,
                                trigger,
                                filter,
                                generic_size,
                                expression_string,
                            )
                    filter["captureFilterExpressionString"] = expression_string
                    trigger["captureTriggerExpressionString"] = (
                        expression_string
                    )
                imports.append(capture)
                imports.append(pallette)
                imports.append(filter)
                imports.append(trigger)
        self._import(imports)

    def reset_capture_request(self):
        self._capture_request = None

    def _config_missing_pallete(
        self,
        cap_filter,
        pallette,
        trigger,
        filter,
        generic_size,
        expression_string,
    ):
        pallete_map = getattr(
            self, "_{0}_OFFSET_MAP".format(cap_filter.parent.choice.upper())
        )
        for field_name in dir(cap_filter):
            if field_name not in pallete_map:
                raise Exception(
                    "Api not implimented for {0}".format(field_name)
                )
            if field_name in pallete_map:
                cap_field = getattr(cap_filter, field_name)
                if cap_field.value is not None:
                    new_pattern = GetPattern(
                        filter.get("captureFilterPattern")
                    )
                    pallette[new_pattern.pattern] = cap_field.value
                    pallette[new_pattern.pattern_offset] = pallete_map[
                        field_name
                    ]
                    if cap_field.mask is not None:
                        pallette[new_pattern.pattern_mask] = cap_field.mask
                    negate = ""
                    if (
                        cap_field.negate is not None
                        and cap_field.negate is True
                    ):
                        filter["captureFilterPattern"] = "notPattern1"
                        negate = "!"
                    else:
                        filter["captureFilterPattern"] = (
                            new_pattern.filter_pattern
                        )
                    trigger["captureTriggerPattern"] = filter[
                        "captureFilterPattern"
                    ]
                    if len(expression_string) == 0:
                        expression_string = negate + "P1"
                    else:
                        expression_string = (
                            expression_string + " and " + negate + "P1"
                        )

        return expression_string

    def _config_custom_pallete(
        self,
        cap_filter,
        pallette,
        trigger,
        filter,
        generic_size,
        expression_string,
    ):
        if cap_filter.value is not None:
            pallette["pattern1"] = cap_filter.value
            if cap_filter.mask is not None:
                pallette["patternMask1"] = cap_filter.mask
            if cap_filter.offset is not None:
                pallette["patternOffset1"] = cap_filter.offset
                generic_size += cap_filter.offset
            negate = ""
            if cap_filter.negate is not None and cap_filter.negate is True:
                filter["captureFilterPattern"] = "notPattern1"
                negate = "!"
            else:
                filter["captureFilterPattern"] = "pattern1"
            trigger["captureTriggerPattern"] = filter["captureFilterPattern"]
            if len(expression_string) == 0:
                expression_string = negate + "P1"
            else:
                expression_string = expression_string + " and " + negate + "P1"

        return generic_size, expression_string

    def _config_ethernet_pallette(
        self,
        cap_filter,
        pallette,
        trigger,
        filter,
        generic_size,
        expression_string,
    ):
        if cap_filter.src.value is not None:
            expression_string = self._config_source_address(
                cap_filter.src, pallette, trigger, filter, expression_string
            )
        if cap_filter.dst.value is not None:
            expression_string = self._config_destination_address(
                cap_filter.dst, pallette, trigger, filter, expression_string
            )

        ethernet_fields = [
            CommonField(cap_filter.ether_type, 12),
            CommonField(cap_filter.pfc_queue, 0),
        ]

        for ethernet_field in ethernet_fields:
            if ethernet_field.field.value != "00":
                expression_string = self._config_common_filter_pallete(
                    ethernet_field.field,
                    pallette,
                    trigger,
                    filter,
                    generic_size + ethernet_field.offset,
                    expression_string,
                )
        return expression_string

    def _config_vlan_pallette(
        self,
        cap_filter,
        pallette,
        trigger,
        filter,
        generic_size,
        expression_string,
    ):
        vlan_fields = [
            CommonField(cap_filter.priority, 0),
            CommonField(cap_filter.cfi, 0),
            CommonField(cap_filter.id, 0),
            CommonField(cap_filter.protocol, 2),
        ]

        for vlan_field in vlan_fields:
            if vlan_field.field.value != "00":
                expression_string = self._config_common_filter_pallete(
                    vlan_field.field,
                    pallette,
                    trigger,
                    filter,
                    generic_size + vlan_field.offset,
                    expression_string,
                )

        return expression_string

    def _config_ipv4_pallete(
        self,
        cap_filter,
        pallette,
        trigger,
        filter,
        generic_size,
        expression_string,
    ):
        ipv4_fields = [
            CommonField(cap_filter.version, 0),
            CommonField(cap_filter.header_length, 0),
            CommonField(cap_filter.priority, 1),
            CommonField(cap_filter.total_length, 2),
            CommonField(cap_filter.identification, 4),
            CommonField(cap_filter.reserved, 6),
            CommonField(cap_filter.dont_fragment, 6),
            CommonField(cap_filter.more_fragments, 6),
            CommonField(cap_filter.fragment_offset, 6),
            CommonField(cap_filter.time_to_live, 8),
            CommonField(cap_filter.protocol, 9),
            CommonField(cap_filter.header_checksum, 10),
            CommonField(cap_filter.src, 12),
            CommonField(cap_filter.dst, 16),
        ]

        for ipv4_field in ipv4_fields:
            if ipv4_field.field.value != "00":
                expression_string = self._config_common_filter_pallete(
                    ipv4_field.field,
                    pallette,
                    trigger,
                    filter,
                    generic_size + ipv4_field.offset,
                    expression_string,
                )

        return expression_string

    def _config_ipv6_pallete(
        self,
        cap_filter,
        pallette,
        trigger,
        filter,
        generic_size,
        expression_string,
    ):
        ipv6_fields = [
            CommonField(cap_filter.version, 0),
            CommonField(cap_filter.traffic_class, 0),
            CommonField(cap_filter.flow_label, 1),
            CommonField(cap_filter.payload_length, 4),
            CommonField(cap_filter.next_header, 6),
            CommonField(cap_filter.hop_limit, 7),
            CommonField(cap_filter.src, 8),
            CommonField(cap_filter.dst, 24),
        ]

        for ipv6_field in ipv6_fields:
            if ipv6_field.field.value != "00":
                expression_string = self._config_common_filter_pallete(
                    ipv6_field.field,
                    pallette,
                    trigger,
                    filter,
                    generic_size + ipv6_field.offset,
                    expression_string,
                )

        return expression_string

    def _hex_to_str_with_space(self, hex_value):
        return " ".join(
            hex_value[i : i + 2] for i in range(0, len(hex_value), 2)
        )

    def _config_source_address(
        self, src, pallette, trigger, filter, expression_string
    ):
        if src.value is not None:
            pallette["SA1"] = self._hex_to_str_with_space(src.value)
            if src.mask is not None:
                pallette["SAMask1"] = self._hex_to_str_with_space(src.mask)
            negate = ""
            if src.negate is not None and src.negate is True:
                filter["captureFilterSA"] = "notAddr1"
                negate = "!"
            else:
                filter["captureFilterSA"] = "addr1"
            trigger["captureTriggerSA"] = filter["captureFilterSA"]
            if len(expression_string) == 0:
                expression_string = negate + "SA1"
            else:
                expression_string = (
                    expression_string + " and " + negate + "SA1"
                )

        return expression_string

    def _config_destination_address(
        self, dst, pallette, trigger, filter, expression_string
    ):
        if dst.value is not None:
            pallette["DA1"] = self._hex_to_str_with_space(dst.value)
            if dst.mask is not None:
                pallette["DAMask1"] = self._hex_to_str_with_space(dst.mask)
            negate = ""
            if dst.negate is not None and dst.negate is True:
                filter["captureFilterDA"] = "notAddr1"
                negate = "!"
            else:
                filter["captureFilterDA"] = "addr1"
            trigger["captureTriggerDA"] = filter["captureFilterDA"]
            if len(expression_string) == 0:
                expression_string = negate + "DA1"
            else:
                expression_string = (
                    expression_string + " and " + negate + "DA1"
                )
        return expression_string

    def _config_common_filter_pallete(
        self, field, pallette, trigger, filter, offset, expression_string
    ):
        if field.value is not None:
            if pallette.get("pattern1") is None:
                pallette["pattern1"] = field.value
                pallette["patternOffset1"] = offset
                if field.mask is not None:
                    pallette["patternMask1"] = field.mask
                negate = ""
                if field.negate is not None and field.negate is True:
                    filter["captureFilterPattern"] = "notPattern1"
                    negate = "!"
                else:
                    filter["captureFilterPattern"] = "pattern1"
                trigger["captureTriggerPattern"] = filter[
                    "captureFilterPattern"
                ]
                if len(expression_string) == 0:
                    expression_string = negate + "P1"
                else:
                    expression_string = (
                        expression_string + " and " + negate + "P1"
                    )
            elif pallette.get("pattern2") is None:
                pallette["pattern2"] = field.value
                pallette["patternOffset2"] = offset
                if field.mask is not None:
                    pallette["patternMask2"] = field.mask
                negate = ""
                if field.negate is not None and field.negate is True:
                    negate = "!"
                filter["captureFilterPattern"] = "pattern1AndPattern2"
                trigger["captureTriggerPattern"] = filter[
                    "captureFilterPattern"
                ]
                if len(expression_string) == 0:
                    expression_string = negate + "P2"
                else:
                    expression_string = (
                        expression_string + " and " + negate + "P2"
                    )
            else:
                self._api.warning("Cannot apply more than 2 filters.")
        return expression_string

    def set_capture_state(self, request):
        """Starts capture on all ports that have capture enabled."""
        self._capture_request = request
        if request.state == "start":
            self._start_capture()
        elif request.state == "stop":
            self._stop_capture()

    def _start_capture(self):
        with Timer(self._api, "Captures start"):
            if self._capture_request is None:
                return
            ixn_vports = self._api.select_vports()
            if len(ixn_vports) == 0:
                raise Exception("Please configure port before start capture")
            ixn_cap_ports = [
                name
                for name, vport in ixn_vports.items()
                if vport["capture"]["hardwareEnabled"] is True
            ]
            port_names = self._capture_request.port_names
            if (
                port_names is None
                or len(set(ixn_cap_ports) ^ set(port_names)) == 0
            ):
                payload = {"arg1": []}
                for vport in ixn_vports.values():
                    payload["arg1"].append(vport["href"])
                url = (
                    "%s/vport/operations/clearCaptureInfos"
                    % self._api._ixnetwork.href
                )
                self._api._request("POST", url, payload)
                self._api._ixnetwork.StartCapture()
            else:
                url = (
                    "%s/vport/capture/operations/start"
                    % self._api._ixnetwork.href
                )
                for vport_name, vport in ixn_vports.items():
                    if vport_name not in port_names:
                        continue
                    if vport["capture"]["hardwareEnabled"] is False:
                        raise Exception(
                            "Please enable capture in"
                            " %s before start capture" % vport_name
                        )
                    payload = {"arg1": "{0}/capture".format(vport["href"])}
                    try:
                        self._api._request("POST", url, payload)
                    except Exception:
                        pass

    def _stop_capture(self):
        with Timer(self._api, "Captures stop"):
            ixn_vports = self._api.select_vports()
            if len(ixn_vports) == 0:
                raise Exception("Please configure port before stop capture")
            if self._capture_request.port_names:
                payload = {"arg1": []}
                for vport_name, vport in ixn_vports.items():
                    if vport_name in self._capture_request.port_names:
                        payload["arg1"].append(vport["href"])
                url = "{}/vport/operations/clearCaptureInfos".format(
                    self._api._ixnetwork.href
                )

                self._api._request("POST", url, payload)
            else:
                self._api._ixnetwork.StopCapture()

    def results(self, request):
        """Gets capture file and returns it as a byte stream"""
        with Timer(self._api, "Captures stop"):
            capture = self._api._vport.find(
                Name=self._api.special_char(request.port_name)
            ).Capture
            capture.Stop("allTraffic")

            # Internally setting max time_out to 90sec,
            #  with 3sec polling interval.
            # Todo: Need to discuss and incorporate time_out field within model
            retry_count = 30
            port_ready = True
            for x in range(retry_count):
                port_ready = True
                time.sleep(3)
                capture = self._api._vport.find(
                    Name=self._api.special_char(request.port_name)
                ).Capture
                if (
                    capture.HardwareEnabled
                    and capture.DataCaptureState == "notReady"
                ):
                    port_ready = False
                    continue
                if (
                    capture.SoftwareEnabled
                    and capture.ControlCaptureState == "notReady"
                ):
                    port_ready = False
                    continue
                break
            if not port_ready:
                self._api.warning(
                    "Capture was not stopped for this port %s"
                    % (request.port_name)
                )

        self._api._ixnetwork.SaveCaptureFiles(
            self._api._ixnetwork.Globals.PersistencePath + "/capture"
        )

        dc = (
            self._api._ixnetwork.Globals.PersistencePath
            + "/capture/"
            + self._api._vport.Name
            + "_HW.cap"
        )
        cc = (
            self._api._ixnetwork.Globals.PersistencePath
            + "/capture/"
            + self._api._vport.Name
            + "_SW.cap"
        )
        merged_capture = (
            self._api._ixnetwork.Globals.PersistencePath
            + "/capture/"
            + self._api._vport.Name
            + ".cap"
        )

        self._api._ixnetwork.MergeCapture(
            Arg1=cc, Arg2=dc, Arg3=merged_capture
        )

        url = "{}/vport/operations/releaseCapturePorts".format(
            self._api._ixnetwork.href
        )
        payload = {"arg1": [self._api._vport.href]}
        self._api._request("POST", url, payload)

        path = "%s/capture" % self._api._ixnetwork.Globals.PersistencePath
        # Todo: Revert dc to merged capture after fix is available in 9.20
        url = "%s/files?absolute=%s&filename=%s" % (
            self._api._ixnetwork.href,
            path,
            dc,
        )
        pcap_file_bytes = self._api._request("GET", url)
        return io.BytesIO(pcap_file_bytes)


class GetPattern(object):
    """This is validating captureFilterPattern and return expected patterns"""

    def __init__(self, cap_filter_pattern):
        self._new_count = 1
        self.cap_filter_pattern = cap_filter_pattern
        self._validate_pattern(cap_filter_pattern)

    def _validate_pattern(self, cap_filter_pattern):
        if cap_filter_pattern is not None:
            self._new_count = 2

    @property
    def pattern(self):
        return "pattern{0}".format(self._new_count)

    @property
    def pattern_mask(self):
        return "patternMask{0}".format(self._new_count)

    @property
    def pattern_offset(self):
        return "patternOffset{0}".format(self._new_count)

    @property
    def filter_pattern(self):
        if self._new_count == 1:
            return "pattern{0}".format(self._new_count)
        else:
            return "{0}AndPattern{1}".format(
                self.cap_filter_pattern, self._new_count
            )
